@@ -1,23 +1,30 @@
-import sys
 import logging
 import weakref
 from time import sleep, time
 from threading import Thread, Event
-from collections import Counter
+from typing import Any, List, Tuple
 
 import pytest
 
 import psycopg
 from psycopg.pq import TransactionStatus
-
-pytestmark = []
+from psycopg._compat import Counter
 
 try:
-    from psycopg_pool import ConnectionPool  # noqa: F401
-except ImportError as ex:
-    pytestmark.append(pytest.mark.skip(reason=str(ex)))
-else:
     import psycopg_pool as pool
+except ImportError:
+    # Tests should have been skipped if the package is not available
+    pass
+
+
+def test_package_version(mypy):
+    cp = mypy.run_on_source(
+        """\
+from psycopg_pool import __version__
+assert __version__
+"""
+    )
+    assert not cp.stdout
 
 
 def test_defaults(dsn):
@@ -29,20 +36,21 @@ def test_defaults(dsn):
         assert p.num_workers == 3
 
 
-def test_min_size_max_size(dsn):
-    with pool.ConnectionPool(dsn, min_size=2) as p:
-        assert p.min_size == p.max_size == 2
+@pytest.mark.parametrize("min_size, max_size", [(2, None), (0, 2), (2, 4)])
+def test_min_size_max_size(dsn, min_size, max_size):
+    with pool.ConnectionPool(dsn, min_size=min_size, max_size=max_size) as p:
+        assert p.min_size == min_size
+        assert p.max_size == max_size if max_size is not None else min_size
 
-    with pool.ConnectionPool(dsn, min_size=2, max_size=4) as p:
-        assert p.min_size == 2
-        assert p.max_size == 4
 
+@pytest.mark.parametrize("min_size, max_size", [(0, 0), (0, None), (-1, None), (4, 2)])
+def test_bad_size(dsn, min_size, max_size):
     with pytest.raises(ValueError):
-        pool.ConnectionPool(dsn, min_size=4, max_size=2)
+        pool.ConnectionPool(min_size=min_size, max_size=max_size)
 
 
 def test_connection_class(dsn):
-    class MyConn(psycopg.Connection):
+    class MyConn(psycopg.Connection[Any]):
         pass
 
     with pool.ConnectionPool(dsn, connection_class=MyConn, min_size=1) as p:
@@ -51,25 +59,22 @@ def test_connection_class(dsn):
 
 
 def test_kwargs(dsn):
-    with pool.ConnectionPool(
-        dsn, kwargs={"autocommit": True}, min_size=1
-    ) as p:
+    with pool.ConnectionPool(dsn, kwargs={"autocommit": True}, min_size=1) as p:
         with p.connection() as conn:
             assert conn.autocommit
 
 
+@pytest.mark.crdb_skip("backend pid")
 def test_its_really_a_pool(dsn):
     with pool.ConnectionPool(dsn, min_size=2) as p:
         with p.connection() as conn:
-            with conn.execute("select pg_backend_pid()") as cur:
-                (pid1,) = cur.fetchone()
+            pid1 = conn.info.backend_pid
 
             with p.connection() as conn2:
-                with conn2.execute("select pg_backend_pid()") as cur:
-                    (pid2,) = cur.fetchone()
+                pid2 = conn2.info.backend_pid
 
         with p.connection() as conn:
-            assert conn.pgconn.backend_pid in (pid1, pid2)
+            assert conn.info.backend_pid in (pid1, pid2)
 
 
 def test_context(dsn):
@@ -78,20 +83,21 @@ def test_context(dsn):
     assert p.closed
 
 
+@pytest.mark.crdb_skip("backend pid")
 def test_connection_not_lost(dsn):
     with pool.ConnectionPool(dsn, min_size=1) as p:
         with pytest.raises(ZeroDivisionError):
             with p.connection() as conn:
-                pid = conn.pgconn.backend_pid
+                pid = conn.info.backend_pid
                 1 / 0
 
         with p.connection() as conn2:
-            assert conn2.pgconn.backend_pid == pid
+            assert conn2.info.backend_pid == pid
 
 
 @pytest.mark.slow
 @pytest.mark.timing
-def test_concurrent_filling(dsn, monkeypatch, retries):
+def test_concurrent_filling(dsn, monkeypatch):
     delay_connection(monkeypatch, 0.1)
 
     def add_time(self, conn):
@@ -101,17 +107,15 @@ def test_concurrent_filling(dsn, monkeypatch, retries):
     add_orig = pool.ConnectionPool._add_to_pool
     monkeypatch.setattr(pool.ConnectionPool, "_add_to_pool", add_time)
 
-    for retry in retries:
-        with retry:
-            times = []
-            t0 = time()
+    times: List[float] = []
+    t0 = time()
 
-            with pool.ConnectionPool(dsn, min_size=5, num_workers=2) as p:
-                p.wait(1.0)
-                want_times = [0.1, 0.1, 0.2, 0.2, 0.3]
-                assert len(times) == len(want_times)
-                for got, want in zip(times, want_times):
-                    assert got == pytest.approx(want, 0.1), times
+    with pool.ConnectionPool(dsn, min_size=5, num_workers=2) as p:
+        p.wait(1.0)
+        want_times = [0.1, 0.1, 0.2, 0.2, 0.3]
+        assert len(times) == len(want_times)
+        for got, want in zip(times, want_times):
+            assert got == pytest.approx(want, 0.1), times
 
 
 @pytest.mark.slow
@@ -130,12 +134,18 @@ def test_wait_ready(dsn, monkeypatch):
         p.wait(0.0001)  # idempotent
 
 
+def test_wait_closed(dsn):
+    with pool.ConnectionPool(dsn) as p:
+        pass
+
+    with pytest.raises(pool.PoolClosed):
+        p.wait()
+
+
 @pytest.mark.slow
 def test_setup_no_timeout(dsn, proxy):
     with pytest.raises(pool.PoolTimeout):
-        with pool.ConnectionPool(
-            proxy.client_dsn, min_size=1, num_workers=1
-        ) as p:
+        with pool.ConnectionPool(proxy.client_dsn, min_size=1, num_workers=1) as p:
             p.wait(0.2)
 
     with pool.ConnectionPool(proxy.client_dsn, min_size=1, num_workers=1) as p:
@@ -157,22 +167,22 @@ def test_configure(dsn):
             conn.execute("set default_transaction_read_only to on")
 
     with pool.ConnectionPool(dsn, min_size=1, configure=configure) as p:
-        p.wait(timeout=1.0)
+        p.wait()
         with p.connection() as conn:
             assert inits == 1
             res = conn.execute("show default_transaction_read_only")
-            assert res.fetchone()[0] == "on"
+            assert res.fetchone()[0] == "on"  # type: ignore[index]
 
         with p.connection() as conn:
             assert inits == 1
             res = conn.execute("show default_transaction_read_only")
-            assert res.fetchone()[0] == "on"
+            assert res.fetchone()[0] == "on"  # type: ignore[index]
             conn.close()
 
         with p.connection() as conn:
             assert inits == 2
             res = conn.execute("show default_transaction_read_only")
-            assert res.fetchone()[0] == "on"
+            assert res.fetchone()[0] == "on"  # type: ignore[index]
 
 
 @pytest.mark.slow
@@ -235,6 +245,7 @@ def test_reset(dsn):
         assert resets == 2
 
 
+@pytest.mark.crdb_skip("backend pid")
 def test_reset_badstate(dsn, caplog):
     caplog.set_level(logging.WARNING, logger="psycopg.pool")
 
@@ -244,17 +255,18 @@ def test_reset_badstate(dsn, caplog):
     with pool.ConnectionPool(dsn, min_size=1, reset=reset) as p:
         with p.connection() as conn:
             conn.execute("select 1")
-            pid1 = conn.pgconn.backend_pid
+            pid1 = conn.info.backend_pid
 
         with p.connection() as conn:
             conn.execute("select 1")
-            pid2 = conn.pgconn.backend_pid
+            pid2 = conn.info.backend_pid
 
     assert pid1 != pid2
     assert caplog.records
     assert "INTRANS" in caplog.records[0].message
 
 
+@pytest.mark.crdb_skip("backend pid")
 def test_reset_broken(dsn, caplog):
     caplog.set_level(logging.WARNING, logger="psycopg.pool")
 
@@ -265,11 +277,11 @@ def test_reset_broken(dsn, caplog):
     with pool.ConnectionPool(dsn, min_size=1, reset=reset) as p:
         with p.connection() as conn:
             conn.execute("select 1")
-            pid1 = conn.pgconn.backend_pid
+            pid1 = conn.info.backend_pid
 
         with p.connection() as conn:
             conn.execute("select 1")
-            pid2 = conn.pgconn.backend_pid
+            pid2 = conn.info.backend_pid
 
     assert pid1 != pid2
     assert caplog.records
@@ -278,31 +290,31 @@ def test_reset_broken(dsn, caplog):
 
 @pytest.mark.slow
 @pytest.mark.timing
-def test_queue(dsn, retries):
+@pytest.mark.crdb_skip("backend pid")
+def test_queue(dsn):
     def worker(n):
         t0 = time()
         with p.connection() as conn:
-            (pid,) = conn.execute(
-                "select pg_backend_pid() from pg_sleep(0.2)"
-            ).fetchone()
+            conn.execute("select pg_sleep(0.2)")
+            pid = conn.info.backend_pid
         t1 = time()
         results.append((n, t1 - t0, pid))
 
-    for retry in retries:
-        with retry:
-            results = []
-            with pool.ConnectionPool(dsn, min_size=2) as p:
-                p.wait()
-                ts = [Thread(target=worker, args=(i,)) for i in range(6)]
-                [t.start() for t in ts]
-                [t.join() for t in ts]
+    results: List[Tuple[int, float, int]] = []
+    with pool.ConnectionPool(dsn, min_size=2) as p:
+        p.wait()
+        ts = [Thread(target=worker, args=(i,)) for i in range(6)]
+        for t in ts:
+            t.start()
+        for t in ts:
+            t.join()
 
-            times = [item[1] for item in results]
-            want_times = [0.2, 0.2, 0.4, 0.4, 0.6, 0.6]
-            for got, want in zip(times, want_times):
-                assert got == pytest.approx(want, 0.1), times
+    times = [item[1] for item in results]
+    want_times = [0.2, 0.2, 0.4, 0.4, 0.6, 0.6]
+    for got, want in zip(times, want_times):
+        assert got == pytest.approx(want, 0.1), times
 
-            assert len(set(r[2] for r in results)) == 2, results
+    assert len(set(r[2] for r in results)) == 2, results
 
 
 @pytest.mark.slow
@@ -318,8 +330,8 @@ def test_queue_size(dsn):
         else:
             success.append(True)
 
-    errors = []
-    success = []
+    errors: List[Exception] = []
+    success: List[bool] = []
 
     with pool.ConnectionPool(dsn, min_size=1, max_waiting=3) as p:
         p.wait()
@@ -329,8 +341,10 @@ def test_queue_size(dsn):
         ev.wait()
 
         ts = [Thread(target=worker, args=(0.1,)) for i in range(4)]
-        [t.start() for t in ts]
-        [t.join() for t in ts]
+        for t in ts:
+            t.start()
+        for t in ts:
+            t.join()
 
     assert len(success) == 4
     assert len(errors) == 1
@@ -342,14 +356,14 @@ def test_queue_size(dsn):
 
 @pytest.mark.slow
 @pytest.mark.timing
-def test_queue_timeout(dsn, retries):
+@pytest.mark.crdb_skip("backend pid")
+def test_queue_timeout(dsn):
     def worker(n):
         t0 = time()
         try:
             with p.connection() as conn:
-                (pid,) = conn.execute(
-                    "select pg_backend_pid() from pg_sleep(0.2)"
-                ).fetchone()
+                conn.execute("select pg_sleep(0.2)")
+                pid = conn.info.backend_pid
         except pool.PoolTimeout as e:
             t1 = time()
             errors.append((n, t1 - t0, e))
@@ -357,20 +371,20 @@ def test_queue_timeout(dsn, retries):
             t1 = time()
             results.append((n, t1 - t0, pid))
 
-    for retry in retries:
-        with retry:
-            results = []
-            errors = []
+    results: List[Tuple[int, float, int]] = []
+    errors: List[Tuple[int, float, Exception]] = []
 
-            with pool.ConnectionPool(dsn, min_size=2, timeout=0.1) as p:
-                ts = [Thread(target=worker, args=(i,)) for i in range(4)]
-                [t.start() for t in ts]
-                [t.join() for t in ts]
+    with pool.ConnectionPool(dsn, min_size=2, timeout=0.1) as p:
+        ts = [Thread(target=worker, args=(i,)) for i in range(4)]
+        for t in ts:
+            t.start()
+        for t in ts:
+            t.join()
 
-            assert len(results) == 2
-            assert len(errors) == 2
-            for e in errors:
-                assert 0.1 < e[1] < 0.15
+    assert len(results) == 2
+    assert len(errors) == 2
+    for e in errors:
+        assert 0.1 < e[1] < 0.15
 
 
 @pytest.mark.slow
@@ -385,15 +399,17 @@ def test_dead_client(dsn):
             if timeout > 0.2:
                 raise
 
-    results = []
+    results: List[int] = []
 
     with pool.ConnectionPool(dsn, min_size=2) as p:
         ts = [
             Thread(target=worker, args=(i, timeout))
             for i, timeout in enumerate([0.4, 0.4, 0.1, 0.4, 0.4])
         ]
-        [t.start() for t in ts]
-        [t.join() for t in ts]
+        for t in ts:
+            t.start()
+        for t in ts:
+            t.join()
         sleep(0.2)
         assert set(results) == set([0, 1, 3, 4])
         assert len(p._pool) == 2  # no connection was lost
@@ -401,15 +417,15 @@ def test_dead_client(dsn):
 
 @pytest.mark.slow
 @pytest.mark.timing
-def test_queue_timeout_override(dsn, retries):
+@pytest.mark.crdb_skip("backend pid")
+def test_queue_timeout_override(dsn):
     def worker(n):
         t0 = time()
         timeout = 0.25 if n == 3 else None
         try:
             with p.connection(timeout=timeout) as conn:
-                (pid,) = conn.execute(
-                    "select pg_backend_pid() from pg_sleep(0.2)"
-                ).fetchone()
+                conn.execute("select pg_sleep(0.2)")
+                pid = conn.info.backend_pid
         except pool.PoolTimeout as e:
             t1 = time()
             errors.append((n, t1 - t0, e))
@@ -417,50 +433,50 @@ def test_queue_timeout_override(dsn, retries):
             t1 = time()
             results.append((n, t1 - t0, pid))
 
-    for retry in retries:
-        with retry:
-            results = []
-            errors = []
+    results: List[Tuple[int, float, int]] = []
+    errors: List[Tuple[int, float, Exception]] = []
 
-            with pool.ConnectionPool(dsn, min_size=2, timeout=0.1) as p:
-                ts = [Thread(target=worker, args=(i,)) for i in range(4)]
-                [t.start() for t in ts]
-                [t.join() for t in ts]
+    with pool.ConnectionPool(dsn, min_size=2, timeout=0.1) as p:
+        ts = [Thread(target=worker, args=(i,)) for i in range(4)]
+        for t in ts:
+            t.start()
+        for t in ts:
+            t.join()
 
-            assert len(results) == 3
-            assert len(errors) == 1
-            for e in errors:
-                assert 0.1 < e[1] < 0.15
+    assert len(results) == 3
+    assert len(errors) == 1
+    for e in errors:
+        assert 0.1 < e[1] < 0.15
 
 
+@pytest.mark.crdb_skip("backend pid")
 def test_broken_reconnect(dsn):
     with pool.ConnectionPool(dsn, min_size=1) as p:
         with p.connection() as conn:
-            with conn.execute("select pg_backend_pid()") as cur:
-                (pid1,) = cur.fetchone()
+            pid1 = conn.info.backend_pid
             conn.close()
 
         with p.connection() as conn2:
-            with conn2.execute("select pg_backend_pid()") as cur:
-                (pid2,) = cur.fetchone()
+            pid2 = conn2.info.backend_pid
 
     assert pid1 != pid2
 
 
+@pytest.mark.crdb_skip("backend pid")
 def test_intrans_rollback(dsn, caplog):
     caplog.set_level(logging.WARNING, logger="psycopg.pool")
 
     with pool.ConnectionPool(dsn, min_size=1) as p:
         conn = p.getconn()
-        pid = conn.pgconn.backend_pid
+        pid = conn.info.backend_pid
         conn.execute("create table test_intrans_rollback ()")
-        assert conn.pgconn.transaction_status == TransactionStatus.INTRANS
+        assert conn.info.transaction_status == TransactionStatus.INTRANS
         p.putconn(conn)
 
         with p.connection() as conn2:
-            assert conn2.pgconn.backend_pid == pid
-            assert conn2.pgconn.transaction_status == TransactionStatus.IDLE
-            assert not conn.execute(
+            assert conn2.info.backend_pid == pid
+            assert conn2.info.transaction_status == TransactionStatus.IDLE
+            assert not conn2.execute(
                 "select 1 from pg_class where relname = 'test_intrans_rollback'"
             ).fetchone()
 
@@ -468,46 +484,48 @@ def test_intrans_rollback(dsn, caplog):
     assert "INTRANS" in caplog.records[0].message
 
 
+@pytest.mark.crdb_skip("backend pid")
 def test_inerror_rollback(dsn, caplog):
     caplog.set_level(logging.WARNING, logger="psycopg.pool")
 
     with pool.ConnectionPool(dsn, min_size=1) as p:
         conn = p.getconn()
-        pid = conn.pgconn.backend_pid
+        pid = conn.info.backend_pid
         with pytest.raises(psycopg.ProgrammingError):
             conn.execute("wat")
-        assert conn.pgconn.transaction_status == TransactionStatus.INERROR
+        assert conn.info.transaction_status == TransactionStatus.INERROR
         p.putconn(conn)
 
         with p.connection() as conn2:
-            assert conn2.pgconn.backend_pid == pid
-            assert conn2.pgconn.transaction_status == TransactionStatus.IDLE
+            assert conn2.info.backend_pid == pid
+            assert conn2.info.transaction_status == TransactionStatus.IDLE
 
     assert len(caplog.records) == 1
     assert "INERROR" in caplog.records[0].message
 
 
+@pytest.mark.crdb_skip("backend pid")
+@pytest.mark.crdb_skip("copy")
 def test_active_close(dsn, caplog):
     caplog.set_level(logging.WARNING, logger="psycopg.pool")
 
     with pool.ConnectionPool(dsn, min_size=1) as p:
         conn = p.getconn()
-        pid = conn.pgconn.backend_pid
-        cur = conn.cursor()
-        with cur.copy("copy (select * from generate_series(1, 10)) to stdout"):
-            pass
-        assert conn.pgconn.transaction_status == TransactionStatus.ACTIVE
+        pid = conn.info.backend_pid
+        conn.pgconn.exec_(b"copy (select * from generate_series(1, 10)) to stdout")
+        assert conn.info.transaction_status == TransactionStatus.ACTIVE
         p.putconn(conn)
 
         with p.connection() as conn2:
-            assert conn2.pgconn.backend_pid != pid
-            assert conn2.pgconn.transaction_status == TransactionStatus.IDLE
+            assert conn2.info.backend_pid != pid
+            assert conn2.info.transaction_status == TransactionStatus.IDLE
 
     assert len(caplog.records) == 2
     assert "ACTIVE" in caplog.records[0].message
     assert "BAD" in caplog.records[1].message
 
 
+@pytest.mark.crdb_skip("backend pid")
 def test_fail_rollback_close(dsn, caplog, monkeypatch):
     caplog.set_level(logging.WARNING, logger="psycopg.pool")
 
@@ -522,15 +540,15 @@ def test_fail_rollback_close(dsn, caplog, monkeypatch):
         orig_rollback = conn.rollback
         monkeypatch.setattr(conn, "rollback", bad_rollback)
 
-        pid = conn.pgconn.backend_pid
+        pid = conn.info.backend_pid
         with pytest.raises(psycopg.ProgrammingError):
             conn.execute("wat")
-        assert conn.pgconn.transaction_status == TransactionStatus.INERROR
+        assert conn.info.transaction_status == TransactionStatus.INERROR
         p.putconn(conn)
 
         with p.connection() as conn2:
-            assert conn2.pgconn.backend_pid != pid
-            assert conn2.pgconn.transaction_status == TransactionStatus.IDLE
+            assert conn2.info.backend_pid != pid
+            assert conn2.info.transaction_status == TransactionStatus.IDLE
 
     assert len(caplog.records) == 3
     assert "INERROR" in caplog.records[0].message
@@ -540,22 +558,27 @@ def test_fail_rollback_close(dsn, caplog, monkeypatch):
 
 def test_close_no_threads(dsn):
     p = pool.ConnectionPool(dsn)
-    assert p._sched_runner.is_alive()
-    for t in p._workers:
+    assert p._sched_runner and p._sched_runner.is_alive()
+    workers = p._workers[:]
+    assert workers
+    for t in workers:
         assert t.is_alive()
 
     p.close()
-    assert not p._sched_runner.is_alive()
-    for t in p._workers:
+    assert p._sched_runner is None
+    assert not p._workers
+    for t in workers:
         assert not t.is_alive()
 
 
-def test_putconn_no_pool(dsn):
+def test_putconn_no_pool(conn_cls, dsn):
     with pool.ConnectionPool(dsn, min_size=1) as p:
-        conn = psycopg.connect(dsn)
+        conn = conn_cls.connect(dsn)
         with pytest.raises(ValueError):
             p.putconn(conn)
 
+    conn.close()
+
 
 def test_putconn_wrong_pool(dsn):
     with pool.ConnectionPool(dsn, min_size=1) as p1:
@@ -580,6 +603,7 @@ def test_del_no_warning(dsn, recwarn):
 @pytest.mark.slow
 def test_del_stop_threads(dsn):
     p = pool.ConnectionPool(dsn)
+    assert p._sched_runner is not None
     ts = [p._sched_runner] + p._workers
     del p
     sleep(0.1)
@@ -613,65 +637,176 @@ def test_closed_putconn(dsn):
     assert conn.closed
 
 
-@pytest.mark.slow
-def test_closed_queue(dsn, retries):
+def test_closed_queue(dsn):
     def w1():
         with p.connection() as conn:
-            assert (
-                conn.execute("select 1 from pg_sleep(0.2)").fetchone()[0] == 1
-            )
+            e1.set()  # Tell w0 that w1 got a connection
+            cur = conn.execute("select 1")
+            assert cur.fetchone() == (1,)
+            e2.wait()  # Wait until w0 has tested w2
         success.append("w1")
 
     def w2():
-        with pytest.raises(pool.PoolClosed):
+        try:
             with p.connection():
-                pass
-        success.append("w2")
+                pass  # unexpected
+        except pool.PoolClosed:
+            success.append("w2")
 
-    for retry in retries:
-        with retry:
-            p = pool.ConnectionPool(dsn, min_size=1)
-            success = []
+    e1 = Event()
+    e2 = Event()
 
-            t1 = Thread(target=w1)
-            t2 = Thread(target=w2)
-            t1.start()
-            sleep(0.1)
-            t2.start()
-            p.close()
-            t1.join()
-            t2.join()
-            assert len(success) == 2
+    p = pool.ConnectionPool(dsn, min_size=1)
+    p.wait()
+    success: List[str] = []
+
+    t1 = Thread(target=w1)
+    t1.start()
+    # Wait until w1 has received a connection
+    e1.wait()
+
+    t2 = Thread(target=w2)
+    t2.start()
+    # Wait until w2 is in the queue
+    ensure_waiting(p)
+
+    p.close(0)
+
+    # Wait for the workers to finish
+    e2.set()
+    t1.join()
+    t2.join()
+    assert len(success) == 2
+
+
+def test_open_explicit(dsn):
+    p = pool.ConnectionPool(dsn, open=False)
+    assert p.closed
+    with pytest.raises(pool.PoolClosed, match="is not open yet"):
+        p.getconn()
+
+    with pytest.raises(pool.PoolClosed):
+        with p.connection():
+            pass
+
+    p.open()
+    try:
+        assert not p.closed
+
+        with p.connection() as conn:
+            cur = conn.execute("select 1")
+            assert cur.fetchone() == (1,)
+
+    finally:
+        p.close()
+
+    with pytest.raises(pool.PoolClosed, match="is already closed"):
+        p.getconn()
+
+
+def test_open_context(dsn):
+    p = pool.ConnectionPool(dsn, open=False)
+    assert p.closed
+
+    with p:
+        assert not p.closed
+
+        with p.connection() as conn:
+            cur = conn.execute("select 1")
+            assert cur.fetchone() == (1,)
+
+    assert p.closed
+
+
+def test_open_no_op(dsn):
+    p = pool.ConnectionPool(dsn)
+    try:
+        assert not p.closed
+        p.open()
+        assert not p.closed
+
+        with p.connection() as conn:
+            cur = conn.execute("select 1")
+            assert cur.fetchone() == (1,)
+
+    finally:
+        p.close()
 
 
 @pytest.mark.slow
 @pytest.mark.timing
-def test_grow(dsn, monkeypatch, retries):
+def test_open_wait(dsn, monkeypatch):
+    delay_connection(monkeypatch, 0.1)
+    with pytest.raises(pool.PoolTimeout):
+        p = pool.ConnectionPool(dsn, min_size=4, num_workers=1, open=False)
+        try:
+            p.open(wait=True, timeout=0.3)
+        finally:
+            p.close()
+
+    p = pool.ConnectionPool(dsn, min_size=4, num_workers=1, open=False)
+    try:
+        p.open(wait=True, timeout=0.5)
+    finally:
+        p.close()
+
+
+@pytest.mark.slow
+@pytest.mark.timing
+def test_open_as_wait(dsn, monkeypatch):
+    delay_connection(monkeypatch, 0.1)
+    with pytest.raises(pool.PoolTimeout):
+        with pool.ConnectionPool(dsn, min_size=4, num_workers=1) as p:
+            p.open(wait=True, timeout=0.3)
+
+    with pool.ConnectionPool(dsn, min_size=4, num_workers=1) as p:
+        p.open(wait=True, timeout=0.5)
+
+
+def test_reopen(dsn):
+    p = pool.ConnectionPool(dsn)
+    with p.connection() as conn:
+        conn.execute("select 1")
+    p.close()
+    assert p._sched_runner is None
+    assert not p._workers
+
+    with pytest.raises(psycopg.OperationalError, match="cannot be reused"):
+        p.open()
+
+
+@pytest.mark.slow
+@pytest.mark.timing
+@pytest.mark.parametrize(
+    "min_size, want_times",
+    [
+        (2, [0.25, 0.25, 0.35, 0.45, 0.50, 0.50, 0.60, 0.70]),
+        (0, [0.35, 0.45, 0.55, 0.60, 0.65, 0.70, 0.80, 0.85]),
+    ],
+)
+def test_grow(dsn, monkeypatch, min_size, want_times):
     delay_connection(monkeypatch, 0.1)
 
     def worker(n):
         t0 = time()
         with p.connection() as conn:
-            conn.execute("select 1 from pg_sleep(0.2)")
+            conn.execute("select 1 from pg_sleep(0.25)")
         t1 = time()
         results.append((n, t1 - t0))
 
-    for retry in retries:
-        with retry:
-            with pool.ConnectionPool(
-                dsn, min_size=2, max_size=4, num_workers=3
-            ) as p:
-                p.wait(1.0)
-                results = []
+    with pool.ConnectionPool(dsn, min_size=min_size, max_size=4, num_workers=3) as p:
+        p.wait(1.0)
+        results: List[Tuple[int, float]] = []
 
-                ts = [Thread(target=worker, args=(i,)) for i in range(6)]
-                [t.start() for t in ts]
-                [t.join() for t in ts]
+        ts = [Thread(target=worker, args=(i,)) for i in range(len(want_times))]
+        for t in ts:
+            t.start()
+        for t in ts:
+            t.join()
 
-            want_times = [0.2, 0.2, 0.3, 0.4, 0.4, 0.4]
-            times = [item[1] for item in results]
-            for got, want in zip(times, want_times):
-                assert got == pytest.approx(want, 0.1), times
+    times = [item[1] for item in results]
+    for got, want in zip(times, want_times):
+        assert got == pytest.approx(want, 0.1), times
 
 
 @pytest.mark.slow
@@ -680,7 +815,7 @@ def test_shrink(dsn, monkeypatch):
 
     from psycopg_pool.pool import ShrinkPool
 
-    results = []
+    results: List[Tuple[int, int]] = []
 
     def run_hacked(self, pool):
         n0 = pool._nconns
@@ -700,15 +835,17 @@ def test_shrink(dsn, monkeypatch):
         assert p.max_idle == 0.2
 
         ts = [Thread(target=worker, args=(i,)) for i in range(4)]
-        [t.start() for t in ts]
-        [t.join() for t in ts]
+        for t in ts:
+            t.start()
+        for t in ts:
+            t.join()
         sleep(1)
 
     assert results == [(4, 4), (4, 3), (3, 2), (2, 2), (2, 2)]
 
 
 @pytest.mark.slow
-def test_reconnect(proxy, caplog, monkeypatch, retries):
+def test_reconnect(proxy, caplog, monkeypatch):
     caplog.set_level(logging.WARNING, logger="psycopg.pool")
 
     assert pool.base.ConnectionAttempt.INITIAL_DELAY == 1.0
@@ -716,36 +853,32 @@ def test_reconnect(proxy, caplog, monkeypatch, retries):
     monkeypatch.setattr(pool.base.ConnectionAttempt, "INITIAL_DELAY", 0.1)
     monkeypatch.setattr(pool.base.ConnectionAttempt, "DELAY_JITTER", 0.0)
 
-    for retry in retries:
-        with retry:
-            caplog.clear()
-            proxy.start()
-            with pool.ConnectionPool(proxy.client_dsn, min_size=1) as p:
-                p.wait(2.0)
-                proxy.stop()
+    caplog.clear()
+    proxy.start()
+    with pool.ConnectionPool(proxy.client_dsn, min_size=1) as p:
+        p.wait(2.0)
+        proxy.stop()
 
-                with pytest.raises(psycopg.OperationalError):
-                    with p.connection() as conn:
-                        conn.execute("select 1")
+        with pytest.raises(psycopg.OperationalError):
+            with p.connection() as conn:
+                conn.execute("select 1")
 
-                sleep(1.0)
-                proxy.start()
-                p.wait()
+        sleep(1.0)
+        proxy.start()
+        p.wait()
 
-                with p.connection() as conn:
-                    conn.execute("select 1")
+        with p.connection() as conn:
+            conn.execute("select 1")
 
-            assert "BAD" in caplog.messages[0]
-            times = [rec.created for rec in caplog.records]
-            assert times[1] - times[0] < 0.05
-            deltas = [
-                times[i + 1] - times[i] for i in range(1, len(times) - 1)
-            ]
-            assert len(deltas) == 3
-            want = 0.1
-            for delta in deltas:
-                assert delta == pytest.approx(want, 0.05), deltas
-                want *= 2
+    assert "BAD" in caplog.messages[0]
+    times = [rec.created for rec in caplog.records]
+    assert times[1] - times[0] < 0.05
+    deltas = [times[i + 1] - times[i] for i in range(1, len(times) - 1)]
+    assert len(deltas) == 3
+    want = 0.1
+    for delta in deltas:
+        assert delta == pytest.approx(want, 0.05), deltas
+        want *= 2
 
 
 @pytest.mark.slow
@@ -789,18 +922,48 @@ def test_reconnect_failure(proxy):
 
 
 @pytest.mark.slow
-def test_uniform_use(dsn, retries):
-    for retry in retries:
-        with retry:
-            with pool.ConnectionPool(dsn, min_size=4) as p:
-                counts = Counter()
-                for i in range(8):
-                    with p.connection() as conn:
-                        sleep(0.1)
-                        counts[id(conn)] += 1
+def test_reconnect_after_grow_failed(proxy):
+    # Retry reconnection after a failed connection attempt has put the pool
+    # in grow mode. See issue #370.
+    proxy.stop()
 
-            assert len(counts) == 4
-            assert set(counts.values()) == set([2])
+    ev = Event()
+
+    def failed(pool):
+        ev.set()
+
+    with pool.ConnectionPool(
+        proxy.client_dsn, min_size=4, reconnect_timeout=1.0, reconnect_failed=failed
+    ) as p:
+        assert ev.wait(timeout=2)
+
+        with pytest.raises(pool.PoolTimeout):
+            with p.connection(timeout=0.5) as conn:
+                pass
+
+        ev.clear()
+        assert ev.wait(timeout=2)
+
+        proxy.start()
+
+        with p.connection(timeout=2) as conn:
+            conn.execute("select 1")
+
+        p.wait(timeout=3.0)
+        assert len(p._pool) == p.min_size == 4
+
+
+@pytest.mark.slow
+def test_uniform_use(dsn):
+    with pool.ConnectionPool(dsn, min_size=4) as p:
+        counts = Counter[int]()
+        for i in range(8):
+            with p.connection() as conn:
+                sleep(0.1)
+                counts[id(conn)] += 1
+
+    assert len(counts) == 4
+    assert set(counts.values()) == set([2])
 
 
 @pytest.mark.slow
@@ -818,7 +981,7 @@ def test_resize(dsn):
         with p.connection() as conn:
             conn.execute("select pg_sleep(%s)", [t])
 
-    size = []
+    size: List[int] = []
 
     with pool.ConnectionPool(dsn, min_size=2, max_idle=0.2) as p:
         s = Thread(target=sampler)
@@ -844,36 +1007,44 @@ def test_resize(dsn):
     assert size == [2, 1, 3, 4, 3, 2, 2]
 
 
+@pytest.mark.parametrize("min_size, max_size", [(0, 0), (-1, None), (4, 2)])
+def test_bad_resize(dsn, min_size, max_size):
+    with pool.ConnectionPool() as p:
+        with pytest.raises(ValueError):
+            p.resize(min_size=min_size, max_size=max_size)
+
+
 def test_jitter():
     rnds = [pool.ConnectionPool._jitter(30, -0.1, +0.2) for i in range(100)]
-    rnds.sort()
     assert 27 <= min(rnds) <= 28
     assert 35 < max(rnds) < 36
 
 
 @pytest.mark.slow
 @pytest.mark.timing
+@pytest.mark.crdb_skip("backend pid")
 def test_max_lifetime(dsn):
     with pool.ConnectionPool(dsn, min_size=1, max_lifetime=0.2) as p:
         sleep(0.1)
         pids = []
         for i in range(5):
             with p.connection() as conn:
-                pids.append(conn.pgconn.backend_pid)
+                pids.append(conn.info.backend_pid)
             sleep(0.2)
 
     assert pids[0] == pids[1] != pids[4], pids
 
 
+@pytest.mark.crdb_skip("backend pid")
 def test_check(dsn, caplog):
     caplog.set_level(logging.WARNING, logger="psycopg.pool")
     with pool.ConnectionPool(dsn, min_size=4) as p:
         p.wait(1.0)
         with p.connection() as conn:
-            pid = conn.pgconn.backend_pid
+            pid = conn.info.backend_pid
 
         p.wait(1.0)
-        pids = set(conn.pgconn.backend_pid for conn in p._pool)
+        pids = set(conn.info.backend_pid for conn in p._pool)
         assert pid in pids
         conn.close()
 
@@ -881,19 +1052,17 @@ def test_check(dsn, caplog):
         p.check()
         assert len(caplog.records) == 1
         p.wait(1.0)
-        pids2 = set(conn.pgconn.backend_pid for conn in p._pool)
+        pids2 = set(conn.info.backend_pid for conn in p._pool)
         assert len(pids & pids2) == 3
         assert pid not in pids2
 
 
-@pytest.mark.skipif(
-    sys.version_info >= (3, 7), reason="async pool supported from Python 3.7"
-)
-def test_async_pool_not_supported(dsn):
-    # note: this test is here because the all the ones in test_pool_async are
-    # skipped on Py 3.6
-    with pytest.raises(psycopg.NotSupportedError):
-        pool.AsyncConnectionPool(dsn)
+def test_check_idle(dsn):
+    with pool.ConnectionPool(dsn, min_size=2) as p:
+        p.wait(1.0)
+        p.check()
+        with p.connection() as conn:
+            assert conn.info.transaction_status == TransactionStatus.IDLE
 
 
 @pytest.mark.slow
@@ -914,10 +1083,12 @@ def test_stats_measures(dsn):
         assert stats["requests_waiting"] == 0
 
         ts = [Thread(target=worker, args=(i,)) for i in range(3)]
-        [t.start() for t in ts]
+        for t in ts:
+            t.start()
         sleep(0.1)
         stats = p.get_stats()
-        [t.join() for t in ts]
+        for t in ts:
+            t.join()
         assert stats["pool_min"] == 2
         assert stats["pool_max"] == 4
         assert stats["pool_size"] == 3
@@ -926,10 +1097,12 @@ def test_stats_measures(dsn):
 
         p.wait(2.0)
         ts = [Thread(target=worker, args=(i,)) for i in range(7)]
-        [t.start() for t in ts]
+        for t in ts:
+            t.start()
         sleep(0.1)
         stats = p.get_stats()
-        [t.join() for t in ts]
+        for t in ts:
+            t.join()
         assert stats["pool_min"] == 2
         assert stats["pool_max"] == 4
         assert stats["pool_size"] == 4
@@ -939,7 +1112,7 @@ def test_stats_measures(dsn):
 
 @pytest.mark.slow
 @pytest.mark.timing
-def test_stats_usage(dsn, retries):
+def test_stats_usage(dsn):
     def worker(n):
         try:
             with p.connection(timeout=0.3) as conn:
@@ -947,31 +1120,31 @@ def test_stats_usage(dsn, retries):
         except pool.PoolTimeout:
             pass
 
-    for retry in retries:
-        with retry:
-            with pool.ConnectionPool(dsn, min_size=3) as p:
-                p.wait(2.0)
+    with pool.ConnectionPool(dsn, min_size=3) as p:
+        p.wait(2.0)
 
-                ts = [Thread(target=worker, args=(i,)) for i in range(7)]
-                [t.start() for t in ts]
-                [t.join() for t in ts]
-                stats = p.get_stats()
-                assert stats["requests_num"] == 7
-                assert stats["requests_queued"] == 4
-                assert 850 <= stats["requests_wait_ms"] <= 950
-                assert stats["requests_errors"] == 1
-                assert 1150 <= stats["usage_ms"] <= 1250
-                assert stats.get("returns_bad", 0) == 0
+        ts = [Thread(target=worker, args=(i,)) for i in range(7)]
+        for t in ts:
+            t.start()
+        for t in ts:
+            t.join()
+        stats = p.get_stats()
+        assert stats["requests_num"] == 7
+        assert stats["requests_queued"] == 4
+        assert 850 <= stats["requests_wait_ms"] <= 950
+        assert stats["requests_errors"] == 1
+        assert 1150 <= stats["usage_ms"] <= 1250
+        assert stats.get("returns_bad", 0) == 0
 
-                with p.connection() as conn:
-                    conn.close()
-                p.wait()
-                stats = p.pop_stats()
-                assert stats["requests_num"] == 8
-                assert stats["returns_bad"] == 1
-                with p.connection():
-                    pass
-                assert p.get_stats()["requests_num"] == 1
+        with p.connection() as conn:
+            conn.close()
+        p.wait()
+        stats = p.pop_stats()
+        assert stats["requests_num"] == 8
+        assert stats["returns_bad"] == 1
+        with p.connection():
+            pass
+        assert p.get_stats()["requests_num"] == 1
 
 
 @pytest.mark.slow
@@ -1009,13 +1182,34 @@ def test_spike(dsn, monkeypatch):
         p.wait()
 
         ts = [Thread(target=worker) for i in range(50)]
-        [t.start() for t in ts]
-        [t.join() for t in ts]
+        for t in ts:
+            t.start()
+        for t in ts:
+            t.join()
         p.wait()
 
         assert len(p._pool) < 7
 
 
+def test_debug_deadlock(dsn):
+    # https://github.com/psycopg/psycopg/issues/230
+    logger = logging.getLogger("psycopg")
+    handler = logging.StreamHandler()
+    old_level = logger.level
+    logger.setLevel(logging.DEBUG)
+    handler.setLevel(logging.DEBUG)
+    logger.addHandler(handler)
+    try:
+        with pool.ConnectionPool(dsn, min_size=4, open=True) as p:
+            try:
+                p.wait(timeout=2)
+            finally:
+                print(p.get_stats())
+    finally:
+        logger.removeHandler(handler)
+        logger.setLevel(old_level)
+
+
 def delay_connection(monkeypatch, sec):
     """
     Return a _connect_gen function delayed by the amount of seconds
@@ -1025,8 +1219,16 @@ def delay_connection(monkeypatch, sec):
         t0 = time()
         rv = connect_orig(*args, **kwargs)
         t1 = time()
-        sleep(sec - (t1 - t0))
+        sleep(max(0, sec - (t1 - t0)))
         return rv
 
     connect_orig = psycopg.Connection.connect
     monkeypatch.setattr(psycopg.Connection, "connect", connect_delay)
+
+
+def ensure_waiting(p, num=1):
+    """
+    Wait until there are at least *num* clients waiting in the queue.
+    """
+    while len(p._waiting) < num:
+        sleep(0)
