@@ -1,7 +1,9 @@
 import gc
 import string
+import struct
 import hashlib
 from io import BytesIO, StringIO
+from random import choice, randrange
 from itertools import cycle
 
 import pytest
@@ -11,42 +13,42 @@ from psycopg import pq
 from psycopg import sql
 from psycopg import errors as e
 from psycopg.pq import Format
+from psycopg.copy import Copy, LibpqWriter, QueuedLibpqDriver, FileWriter
 from psycopg.adapt import PyFormat
 from psycopg.types import TypeInfo
 from psycopg.types.hstore import register_hstore
 from psycopg.types.numeric import Int4
 
-from .utils import gc_collect
+from .utils import eur, gc_collect
 
-eur = "\u20ac"
-
-sample_records = [(10, 20, "hello"), (40, None, "world")]
-
-sample_values = "values (10::int, 20::int, 'hello'::text), (40, NULL, 'world')"
+pytestmark = pytest.mark.crdb_skip("copy")
 
+sample_records = [(40010, 40020, "hello"), (40040, None, "world")]
+sample_values = "values (40010::int, 40020::int, 'hello'::text), (40040, NULL, 'world')"
 sample_tabledef = "col1 serial primary key, col2 int, data text"
 
 sample_text = b"""\
-10\t20\thello
-40\t\\N\tworld
+40010\t40020\thello
+40040\t\\N\tworld
 """
 
-sample_binary = """
+sample_binary_str = """
 5047 434f 5059 0aff 0d0a 00
 00 0000 0000 0000 00
-00 0300 0000 0400 0000 0a00 0000 0400 0000 1400 0000 0568 656c 6c6f
+00 0300 0000 0400 009c 4a00 0000 0400 009c 5400 0000 0568 656c 6c6f
 
-0003 0000 0004 0000 0028 ffff ffff 0000 0005 776f 726c 64
+0003 0000 0004 0000 9c68 ffff ffff 0000 0005 776f 726c 64
 
 ff ff
 """
 
 sample_binary_rows = [
-    bytes.fromhex("".join(row.split())) for row in sample_binary.split("\n\n")
+    bytes.fromhex("".join(row.split())) for row in sample_binary_str.split("\n\n")
 ]
-
 sample_binary = b"".join(sample_binary_rows)
 
+special_chars = {8: "b", 9: "t", 10: "n", 11: "v", 12: "f", 13: "r", ord("\\"): "\\"}
+
 
 @pytest.mark.parametrize("format", Format)
 def test_copy_out_read(conn, format):
@@ -56,21 +58,17 @@ def test_copy_out_read(conn, format):
         want = sample_binary_rows
 
     cur = conn.cursor()
-    with cur.copy(
-        f"copy ({sample_values}) to stdout (format {format.name})"
-    ) as copy:
+    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
         for row in want:
             got = copy.read()
             assert got == row
-            assert (
-                conn.pgconn.transaction_status == conn.TransactionStatus.ACTIVE
-            )
+            assert conn.info.transaction_status == conn.TransactionStatus.ACTIVE
 
         assert copy.read() == b""
         assert copy.read() == b""
 
     assert copy.read() == b""
-    assert conn.pgconn.transaction_status == conn.TransactionStatus.INTRANS
+    assert conn.info.transaction_status == conn.TransactionStatus.INTRANS
 
 
 @pytest.mark.parametrize("format", Format)
@@ -81,12 +79,22 @@ def test_copy_out_iter(conn, format):
         want = sample_binary_rows
 
     cur = conn.cursor()
-    with cur.copy(
-        f"copy ({sample_values}) to stdout (format {format.name})"
-    ) as copy:
+    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
         assert list(copy) == want
 
-    assert conn.pgconn.transaction_status == conn.TransactionStatus.INTRANS
+    assert conn.info.transaction_status == conn.TransactionStatus.INTRANS
+
+
+@pytest.mark.parametrize("ph, params", [("%s", (10,)), ("%(n)s", {"n": 10})])
+def test_copy_out_param(conn, ph, params):
+    cur = conn.cursor()
+    with cur.copy(
+        f"copy (select * from generate_series(1, {ph})) to stdout", params
+    ) as copy:
+        copy.set_types(["int4"])
+        assert list(copy.rows()) == [(i + 1,) for i in range(10)]
+
+    assert conn.info.transaction_status == conn.TransactionStatus.INTRANS
 
 
 @pytest.mark.parametrize("format", Format)
@@ -103,20 +111,18 @@ def test_read_rows(conn, format, typetype):
         assert copy.read_row() is None
 
     assert row == (10, "hello", [0.0, 1.0])
-    assert conn.pgconn.transaction_status == conn.TransactionStatus.INTRANS
+    assert conn.info.transaction_status == conn.TransactionStatus.INTRANS
 
 
 @pytest.mark.parametrize("format", Format)
 def test_rows(conn, format):
     cur = conn.cursor()
-    with cur.copy(
-        f"copy ({sample_values}) to stdout (format {format.name})"
-    ) as copy:
+    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
         copy.set_types(["int4", "int4", "text"])
         rows = list(copy.rows())
 
     assert rows == sample_records
-    assert conn.pgconn.transaction_status == conn.TransactionStatus.INTRANS
+    assert conn.info.transaction_status == conn.TransactionStatus.INTRANS
 
 
 def test_set_custom_type(conn, hstore):
@@ -142,12 +148,12 @@ def test_copy_out_allchars(conn, format):
     chars = list(map(chr, range(1, 256))) + [eur]
     conn.execute("set client_encoding to utf8")
     rows = []
-    query = sql.SQL(
-        "copy (select unnest({}::text[])) to stdout (format {})"
-    ).format(chars, sql.SQL(format.name))
+    query = sql.SQL("copy (select unnest({}::text[])) to stdout (format {})").format(
+        chars, sql.SQL(format.name)
+    )
     with cur.copy(query) as copy:
         copy.set_types(["text"])
-        while 1:
+        while True:
             row = copy.read_row()
             if not row:
                 break
@@ -160,34 +166,24 @@ def test_copy_out_allchars(conn, format):
 @pytest.mark.parametrize("format", Format)
 def test_read_row_notypes(conn, format):
     cur = conn.cursor()
-    with cur.copy(
-        f"copy ({sample_values}) to stdout (format {format.name})"
-    ) as copy:
+    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
         rows = []
-        while 1:
+        while True:
             row = copy.read_row()
             if not row:
                 break
             rows.append(row)
 
-    ref = [
-        tuple(py_to_raw(i, format) for i in record)
-        for record in sample_records
-    ]
+    ref = [tuple(py_to_raw(i, format) for i in record) for record in sample_records]
     assert rows == ref
 
 
 @pytest.mark.parametrize("format", Format)
 def test_rows_notypes(conn, format):
     cur = conn.cursor()
-    with cur.copy(
-        f"copy ({sample_values}) to stdout (format {format.name})"
-    ) as copy:
+    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
         rows = list(copy.rows())
-    ref = [
-        tuple(py_to_raw(i, format) for i in record)
-        for record in sample_records
-    ]
+    ref = [tuple(py_to_raw(i, format) for i in record) for record in sample_records]
     assert rows == ref
 
 
@@ -195,9 +191,7 @@ def test_rows_notypes(conn, format):
 @pytest.mark.parametrize("format", Format)
 def test_copy_out_badntypes(conn, format, err):
     cur = conn.cursor()
-    with cur.copy(
-        f"copy ({sample_values}) to stdout (format {format.name})"
-    ) as copy:
+    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
         copy.set_types([0] * (len(sample_records[0]) + err))
         with pytest.raises(e.ProgrammingError):
             copy.read_row()
@@ -224,7 +218,7 @@ def test_copy_in_buffers_pg_error(conn):
         with cur.copy("copy copy_in from stdin (format text)") as copy:
             copy.write(sample_text)
             copy.write(sample_text)
-    assert conn.pgconn.transaction_status == conn.TransactionStatus.INERROR
+    assert conn.info.transaction_status == conn.TransactionStatus.INERROR
 
 
 def test_copy_bad_result(conn):
@@ -263,14 +257,14 @@ def test_copy_in_str(conn):
     assert data == sample_records
 
 
-def test_copy_in_str_binary(conn):
+def test_copy_in_error(conn):
     cur = conn.cursor()
     ensure_table(cur, sample_tabledef)
     with pytest.raises(e.QueryCanceled):
         with cur.copy("copy copy_in from stdin (format binary)") as copy:
             copy.write(sample_text.decode())
 
-    assert conn.pgconn.transaction_status == conn.TransactionStatus.INERROR
+    assert conn.info.transaction_status == conn.TransactionStatus.INERROR
 
 
 @pytest.mark.parametrize("format", Format)
@@ -280,16 +274,44 @@ def test_copy_in_empty(conn, format):
     with cur.copy(f"copy copy_in from stdin (format {format.name})"):
         pass
 
-    assert conn.pgconn.transaction_status == conn.TransactionStatus.INTRANS
+    assert conn.info.transaction_status == conn.TransactionStatus.INTRANS
     assert cur.rowcount == 0
 
 
+@pytest.mark.slow
+def test_copy_big_size_record(conn):
+    cur = conn.cursor()
+    ensure_table(cur, sample_tabledef)
+    data = "".join(chr(randrange(1, 256)) for i in range(10 * 1024 * 1024))
+    with cur.copy("copy copy_in (data) from stdin") as copy:
+        copy.write_row([data])
+
+    cur.execute("select data from copy_in limit 1")
+    assert cur.fetchone()[0] == data
+
+
+@pytest.mark.slow
+@pytest.mark.parametrize("pytype", [str, bytes, bytearray, memoryview])
+def test_copy_big_size_block(conn, pytype):
+    cur = conn.cursor()
+    ensure_table(cur, sample_tabledef)
+    data = "".join(choice(string.ascii_letters) for i in range(10 * 1024 * 1024))
+    copy_data = data + "\n" if pytype is str else pytype(data.encode() + b"\n")
+    with cur.copy("copy copy_in (data) from stdin") as copy:
+        copy.write(copy_data)
+
+    cur.execute("select data from copy_in limit 1")
+    assert cur.fetchone()[0] == data
+
+
 @pytest.mark.parametrize("format", Format)
 def test_subclass_adapter(conn, format):
     if format == Format.TEXT:
         from psycopg.types.string import StrDumper as BaseDumper
     else:
-        from psycopg.types.string import StrBinaryDumper as BaseDumper
+        from psycopg.types.string import (  # type: ignore[no-redef]
+            StrBinaryDumper as BaseDumper,
+        )
 
     class MyStrDumper(BaseDumper):
         def dump(self, obj):
@@ -300,9 +322,7 @@ def test_subclass_adapter(conn, format):
     cur = conn.cursor()
     ensure_table(cur, sample_tabledef)
 
-    with cur.copy(
-        f"copy copy_in (data) from stdin (format {format.name})"
-    ) as copy:
+    with cur.copy(f"copy copy_in (data) from stdin (format {format.name})") as copy:
         copy.write_row(("hello",))
 
     rec = cur.execute("select data from copy_in").fetchone()
@@ -318,7 +338,7 @@ def test_copy_in_error_empty(conn, format):
             raise Exception("mannaggiamiseria")
 
     assert "mannaggiamiseria" in str(exc.value)
-    assert conn.pgconn.transaction_status == conn.TransactionStatus.INERROR
+    assert conn.info.transaction_status == conn.TransactionStatus.INERROR
 
 
 def test_copy_in_buffers_with_pg_error(conn):
@@ -329,7 +349,7 @@ def test_copy_in_buffers_with_pg_error(conn):
             copy.write(sample_text)
             copy.write(sample_text)
 
-    assert conn.pgconn.transaction_status == conn.TransactionStatus.INERROR
+    assert conn.info.transaction_status == conn.TransactionStatus.INERROR
 
 
 def test_copy_in_buffers_with_py_error(conn):
@@ -341,7 +361,39 @@ def test_copy_in_buffers_with_py_error(conn):
             raise Exception("nuttengoggenio")
 
     assert "nuttengoggenio" in str(exc.value)
-    assert conn.pgconn.transaction_status == conn.TransactionStatus.INERROR
+    assert conn.info.transaction_status == conn.TransactionStatus.INERROR
+
+
+def test_copy_out_error_with_copy_finished(conn):
+    cur = conn.cursor()
+    with pytest.raises(ZeroDivisionError):
+        with cur.copy("copy (select generate_series(1, 2)) to stdout") as copy:
+            copy.read_row()
+            1 / 0
+
+    assert conn.info.transaction_status == conn.TransactionStatus.INTRANS
+
+
+def test_copy_out_error_with_copy_not_finished(conn):
+    cur = conn.cursor()
+    with pytest.raises(ZeroDivisionError):
+        with cur.copy("copy (select generate_series(1, 1000000)) to stdout") as copy:
+            copy.read_row()
+            1 / 0
+
+    assert conn.info.transaction_status == conn.TransactionStatus.INERROR
+
+
+def test_copy_out_server_error(conn):
+    cur = conn.cursor()
+    with pytest.raises(e.DivisionByZero):
+        with cur.copy(
+            "copy (select 1/n from generate_series(-10, 10) x(n)) to stdout"
+        ) as copy:
+            for block in copy:
+                pass
+
+    assert conn.info.transaction_status == conn.TransactionStatus.INERROR
 
 
 @pytest.mark.parametrize("format", Format)
@@ -352,7 +404,9 @@ def test_copy_in_records(conn, format):
     with cur.copy(f"copy copy_in from stdin (format {format.name})") as copy:
         for row in sample_records:
             if format == Format.BINARY:
-                row = tuple(Int4(i) if isinstance(i, int) else i for i in row)
+                row = tuple(
+                    Int4(i) if isinstance(i, int) else i for i in row
+                )  # type: ignore[assignment]
             copy.write_row(row)
 
     data = cur.execute("select * from copy_in order by 1").fetchall()
@@ -407,6 +461,46 @@ from copy_in group by 1, 2, 3
     assert data == [(True, True, 1, 256)]
 
 
+def test_copy_in_format(conn):
+    file = BytesIO()
+    conn.execute("set client_encoding to utf8")
+    cur = conn.cursor()
+    with Copy(cur, writer=FileWriter(file)) as copy:
+        for i in range(1, 256):
+            copy.write_row((i, chr(i)))
+
+    file.seek(0)
+    rows = file.read().split(b"\n")
+    assert not rows[-1]
+    del rows[-1]
+
+    for i, row in enumerate(rows, start=1):
+        fields = row.split(b"\t")
+        assert len(fields) == 2
+        assert int(fields[0].decode()) == i
+        if i in special_chars:
+            assert fields[1].decode() == f"\\{special_chars[i]}"
+        else:
+            assert fields[1].decode() == chr(i)
+
+
+@pytest.mark.parametrize(
+    "format, buffer", [(Format.TEXT, "sample_text"), (Format.BINARY, "sample_binary")]
+)
+def test_file_writer(conn, format, buffer):
+    file = BytesIO()
+    conn.execute("set client_encoding to utf8")
+    cur = conn.cursor()
+    with Copy(cur, binary=format, writer=FileWriter(file)) as copy:
+        for record in sample_records:
+            copy.write_row(record)
+
+    file.seek(0)
+    want = globals()[buffer]
+    got = file.read()
+    assert got == want
+
+
 @pytest.mark.slow
 def test_copy_from_to(conn):
     # Roundtrip from file to database to file blockwise
@@ -429,14 +523,15 @@ def test_copy_from_to(conn):
 
 
 @pytest.mark.slow
-def test_copy_from_to_bytes(conn):
+@pytest.mark.parametrize("pytype", [bytes, bytearray, memoryview])
+def test_copy_from_to_bytes(conn, pytype):
     # Roundtrip from file to database to file blockwise
     gen = DataGenerator(conn, nrecs=1024, srec=10 * 1024)
     gen.ensure_table()
     cur = conn.cursor()
     with cur.copy("copy copy_in from stdin") as copy:
         for block in gen.blocks():
-            copy.write(block.encode())
+            copy.write(pytype(block.encode()))
 
     gen.assert_data()
 
@@ -519,19 +614,64 @@ def test_str(conn):
     assert "[INTRANS]" in str(copy)
 
 
+def test_description(conn):
+    with conn.cursor() as cur:
+        with cur.copy("copy (select 'This', 'Is', 'Text') to stdout") as copy:
+            len(cur.description) == 3
+            assert cur.description[0].name == "column_1"
+            assert cur.description[2].name == "column_3"
+            list(copy.rows())
+
+        len(cur.description) == 3
+        assert cur.description[0].name == "column_1"
+        assert cur.description[2].name == "column_3"
+
+
 @pytest.mark.parametrize(
-    "format, buffer",
-    [(Format.TEXT, "sample_text"), (Format.BINARY, "sample_binary")],
+    "format, buffer", [(Format.TEXT, "sample_text"), (Format.BINARY, "sample_binary")]
 )
 def test_worker_life(conn, format, buffer):
     cur = conn.cursor()
     ensure_table(cur, sample_tabledef)
-    with cur.copy(f"copy copy_in from stdin (format {format.name})") as copy:
-        assert not copy._worker
+    with cur.copy(
+        f"copy copy_in from stdin (format {format.name})", writer=QueuedLibpqDriver(cur)
+    ) as copy:
+        assert not copy.writer._worker
+        copy.write(globals()[buffer])
+        assert copy.writer._worker
+
+    assert not copy.writer._worker
+    data = cur.execute("select * from copy_in order by 1").fetchall()
+    assert data == sample_records
+
+
+def test_worker_error_propagated(conn, monkeypatch):
+    def copy_to_broken(pgconn, buffer):
+        raise ZeroDivisionError
+        yield
+
+    monkeypatch.setattr(psycopg.copy, "copy_to", copy_to_broken)
+    cur = conn.cursor()
+    cur.execute("create temp table wat (a text, b text)")
+    with pytest.raises(ZeroDivisionError):
+        with cur.copy("copy wat from stdin", writer=QueuedLibpqDriver(cur)) as copy:
+            copy.write("a,b")
+
+
+@pytest.mark.parametrize(
+    "format, buffer", [(Format.TEXT, "sample_text"), (Format.BINARY, "sample_binary")]
+)
+def test_connection_writer(conn, format, buffer):
+    cur = conn.cursor()
+    writer = LibpqWriter(cur)
+
+    ensure_table(cur, sample_tabledef)
+    with cur.copy(
+        f"copy copy_in from stdin (format {format.name})", writer=writer
+    ) as copy:
+        assert copy.writer is writer
         copy.write(globals()[buffer])
-        assert copy._worker
 
-    assert not copy._worker
     data = cur.execute("select * from copy_in order by 1").fetchall()
     assert data == sample_records
 
@@ -542,13 +682,13 @@ def test_worker_life(conn, format, buffer):
     [(Format.TEXT, True), (Format.TEXT, False), (Format.BINARY, True)],
 )
 @pytest.mark.parametrize("method", ["read", "iter", "row", "rows"])
-def test_copy_to_leaks(dsn, faker, fmt, set_types, method, retries):
+def test_copy_to_leaks(conn_cls, dsn, faker, fmt, set_types, method):
     faker.format = PyFormat.from_pq(fmt)
     faker.choose_schema(ncols=20)
     faker.make_records(20)
 
     def work():
-        with psycopg.connect(dsn) as conn:
+        with conn_cls.connect(dsn) as conn:
             with conn.cursor(binary=fmt) as cur:
                 cur.execute(faker.drop_stmt)
                 cur.execute(faker.create_stmt)
@@ -568,14 +708,14 @@ def test_copy_to_leaks(dsn, faker, fmt, set_types, method, retries):
                         copy.set_types(faker.types_names)
 
                     if method == "read":
-                        while 1:
+                        while True:
                             tmp = copy.read()
                             if not tmp:
                                 break
                     elif method == "iter":
                         list(copy)
                     elif method == "row":
-                        while 1:
+                        while True:
                             tmp = copy.read_row()
                             if tmp is None:
                                 break
@@ -583,17 +723,13 @@ def test_copy_to_leaks(dsn, faker, fmt, set_types, method, retries):
                         list(copy.rows())
 
     gc_collect()
-    for retry in retries:
-        with retry:
-            n = []
-            for i in range(3):
-                work()
-                gc_collect()
-                n.append(len(gc.get_objects()))
+    n = []
+    for i in range(3):
+        work()
+        gc_collect()
+        n.append(len(gc.get_objects()))
 
-            assert (
-                n[0] == n[1] == n[2]
-            ), f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"
+    assert n[0] == n[1] == n[2], f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"
 
 
 @pytest.mark.slow
@@ -601,13 +737,13 @@ def test_copy_to_leaks(dsn, faker, fmt, set_types, method, retries):
     "fmt, set_types",
     [(Format.TEXT, True), (Format.TEXT, False), (Format.BINARY, True)],
 )
-def test_copy_from_leaks(dsn, faker, fmt, set_types, retries):
+def test_copy_from_leaks(conn_cls, dsn, faker, fmt, set_types):
     faker.format = PyFormat.from_pq(fmt)
     faker.choose_schema(ncols=20)
     faker.make_records(20)
 
     def work():
-        with psycopg.connect(dsn) as conn:
+        with conn_cls.connect(dsn) as conn:
             with conn.cursor(binary=fmt) as cur:
                 cur.execute(faker.drop_stmt)
                 cur.execute(faker.create_stmt)
@@ -630,17 +766,44 @@ def test_copy_from_leaks(dsn, faker, fmt, set_types, retries):
                     faker.assert_record(got, want)
 
     gc_collect()
-    for retry in retries:
-        with retry:
-            n = []
-            for i in range(3):
-                work()
-                gc_collect()
-                n.append(len(gc.get_objects()))
+    n = []
+    for i in range(3):
+        work()
+        gc_collect()
+        n.append(len(gc.get_objects()))
 
-            assert (
-                n[0] == n[1] == n[2]
-            ), f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"
+    assert n[0] == n[1] == n[2], f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"
+
+
+@pytest.mark.slow
+@pytest.mark.parametrize("mode", ["row", "block", "binary"])
+def test_copy_table_across(conn_cls, dsn, faker, mode):
+    faker.choose_schema(ncols=20)
+    faker.make_records(20)
+
+    with conn_cls.connect(dsn) as conn1, conn_cls.connect(dsn) as conn2:
+        faker.table_name = sql.Identifier("copy_src")
+        conn1.execute(faker.drop_stmt)
+        conn1.execute(faker.create_stmt)
+        conn1.cursor().executemany(faker.insert_stmt, faker.records)
+
+        faker.table_name = sql.Identifier("copy_tgt")
+        conn2.execute(faker.drop_stmt)
+        conn2.execute(faker.create_stmt)
+
+        fmt = "(format binary)" if mode == "binary" else ""
+        with conn1.cursor().copy(f"copy copy_src to stdout {fmt}") as copy1:
+            with conn2.cursor().copy(f"copy copy_tgt from stdin {fmt}") as copy2:
+                if mode == "row":
+                    for row in copy1.rows():
+                        copy2.write_row(row)
+                else:
+                    for data in copy1:
+                        copy2.write(data)
+
+        recs = conn2.execute(faker.select_stmt).fetchall()
+        for got, want in zip(recs, faker.records):
+            faker.assert_record(got, want)
 
 
 def py_to_raw(item, fmt):
@@ -650,7 +813,8 @@ def py_to_raw(item, fmt):
             return str(item)
     else:
         if isinstance(item, int):
-            return bytes([0, 0, 0, item])
+            # Assume int4
+            return struct.pack("!i", item)
         elif isinstance(item, str):
             return item.encode()
     return item
@@ -704,7 +868,7 @@ class DataGenerator:
 
     def sha(self, f):
         m = hashlib.sha256()
-        while 1:
+        while True:
             block = f.read()
             if not block:
                 break
