@@ -2,6 +2,8 @@ import gc
 import pickle
 import weakref
 import datetime as dt
+from typing import List, Union
+from contextlib import closing
 
 import pytest
 
@@ -9,8 +11,27 @@ import psycopg
 from psycopg import pq, sql, rows
 from psycopg.adapt import PyFormat
 from psycopg.postgres import types as builtins
+from psycopg.rows import RowMaker
 
 from .utils import gc_collect
+from .fix_crdb import is_crdb, crdb_encoding, crdb_time_precision
+
+
+def test_init(conn):
+    cur = psycopg.Cursor(conn)
+    cur.execute("select 1")
+    assert cur.fetchone() == (1,)
+
+    conn.row_factory = rows.dict_row
+    cur = psycopg.Cursor(conn)
+    cur.execute("select 1 as a")
+    assert cur.fetchone() == {"a": 1}
+
+
+def test_init_factory(conn):
+    cur = psycopg.Cursor(conn, row_factory=rows.dict_row)
+    cur.execute("select 1 as a")
+    assert cur.fetchone() == {"a": 1}
 
 
 def test_close(conn):
@@ -26,6 +47,52 @@ def test_close(conn):
     assert cur.closed
 
 
+def test_cursor_close_fetchone(conn):
+    cur = conn.cursor()
+    assert not cur.closed
+
+    query = "select * from generate_series(1, 10)"
+    cur.execute(query)
+    for _ in range(5):
+        cur.fetchone()
+
+    cur.close()
+    assert cur.closed
+
+    with pytest.raises(psycopg.InterfaceError):
+        cur.fetchone()
+
+
+def test_cursor_close_fetchmany(conn):
+    cur = conn.cursor()
+    assert not cur.closed
+
+    query = "select * from generate_series(1, 10)"
+    cur.execute(query)
+    assert len(cur.fetchmany(2)) == 2
+
+    cur.close()
+    assert cur.closed
+
+    with pytest.raises(psycopg.InterfaceError):
+        cur.fetchmany(2)
+
+
+def test_cursor_close_fetchall(conn):
+    cur = conn.cursor()
+    assert not cur.closed
+
+    query = "select * from generate_series(1, 10)"
+    cur.execute(query)
+    assert len(cur.fetchall()) == 10
+
+    cur.close()
+    assert cur.closed
+
+    with pytest.raises(psycopg.InterfaceError):
+        cur.fetchall()
+
+
 def test_context(conn):
     with conn.cursor() as cur:
         assert not cur.closed
@@ -33,6 +100,7 @@ def test_context(conn):
     assert cur.closed
 
 
+@pytest.mark.slow
 def test_weakref(conn):
     cur = conn.cursor()
     w = weakref.ref(cur)
@@ -42,6 +110,14 @@ def test_weakref(conn):
     assert w() is None
 
 
+def test_pgresult(conn):
+    cur = conn.cursor()
+    cur.execute("select 1")
+    assert cur.pgresult
+    cur.close()
+    assert not cur.pgresult
+
+
 def test_statusmessage(conn):
     cur = conn.cursor()
     assert cur.statusmessage is None
@@ -158,7 +234,7 @@ def test_binary_cursor_text_override(conn):
     assert cur.pgresult.get_value(0, 0) == b"1"
 
 
-@pytest.mark.parametrize("encoding", ["utf8", "latin9"])
+@pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
 def test_query_encode(conn, encoding):
     conn.execute(f"set client_encoding to {encoding}")
     cur = conn.cursor()
@@ -166,8 +242,9 @@ def test_query_encode(conn, encoding):
     assert res == "\u20ac"
 
 
-def test_query_badenc(conn):
-    conn.execute("set client_encoding to latin1")
+@pytest.mark.parametrize("encoding", [crdb_encoding("latin1")])
+def test_query_badenc(conn, encoding):
+    conn.execute(f"set client_encoding to {encoding}")
     cur = conn.cursor()
     with pytest.raises(UnicodeEncodeError):
         cur.execute("select '\u20ac'")
@@ -210,6 +287,12 @@ def test_executemany_name(conn, execmany):
     assert cur.fetchall() == [(11, "hello"), (21, "world")]
 
 
+def test_executemany_no_data(conn, execmany):
+    cur = conn.cursor()
+    cur.executemany("insert into execmany(num, data) values (%s, %s)", [])
+    assert cur.rowcount == 0
+
+
 def test_executemany_rowcount(conn, execmany):
     cur = conn.cursor()
     cur.executemany(
@@ -219,13 +302,58 @@ def test_executemany_rowcount(conn, execmany):
     assert cur.rowcount == 2
 
 
-def test_executemany_returning_rowcount(conn, execmany):
+def test_executemany_returning(conn, execmany):
+    cur = conn.cursor()
+    cur.executemany(
+        "insert into execmany(num, data) values (%s, %s) returning num",
+        [(10, "hello"), (20, "world")],
+        returning=True,
+    )
+    assert cur.rowcount == 2
+    assert cur.fetchone() == (10,)
+    assert cur.nextset()
+    assert cur.fetchone() == (20,)
+    assert cur.nextset() is None
+
+
+def test_executemany_returning_discard(conn, execmany):
     cur = conn.cursor()
     cur.executemany(
         "insert into execmany(num, data) values (%s, %s) returning num",
         [(10, "hello"), (20, "world")],
     )
     assert cur.rowcount == 2
+    with pytest.raises(psycopg.ProgrammingError):
+        cur.fetchone()
+    assert cur.nextset() is None
+
+
+def test_executemany_no_result(conn, execmany):
+    cur = conn.cursor()
+    cur.executemany(
+        "insert into execmany(num, data) values (%s, %s)",
+        [(10, "hello"), (20, "world")],
+        returning=True,
+    )
+    assert cur.rowcount == 2
+    assert cur.statusmessage.startswith("INSERT")
+    with pytest.raises(psycopg.ProgrammingError):
+        cur.fetchone()
+    pgresult = cur.pgresult
+    assert cur.nextset()
+    assert cur.statusmessage.startswith("INSERT")
+    assert pgresult is not cur.pgresult
+    assert cur.nextset() is None
+
+
+def test_executemany_rowcount_no_hit(conn, execmany):
+    cur = conn.cursor()
+    cur.executemany("delete from execmany where id = %s", [(-1,), (-2,)])
+    assert cur.rowcount == 0
+    cur.executemany("delete from execmany where id = %s", [])
+    assert cur.rowcount == 0
+    cur.executemany("delete from execmany where id = %s returning num", [(-1,), (-2,)])
+    assert cur.rowcount == 0
 
 
 @pytest.mark.parametrize(
@@ -247,12 +375,12 @@ def test_executemany_null_first(conn, fmt_in):
     cur = conn.cursor()
     cur.execute("create table testmany (a bigint, b bigint)")
     cur.executemany(
-        f"insert into testmany values (%{fmt_in}, %{fmt_in})",
+        f"insert into testmany values (%{fmt_in.value}, %{fmt_in.value})",
         [[1, None], [3, 4]],
     )
     with pytest.raises((psycopg.DataError, psycopg.ProgrammingError)):
         cur.executemany(
-            f"insert into testmany values (%{fmt_in}, %{fmt_in})",
+            f"insert into testmany values (%{fmt_in.value}, %{fmt_in.value})",
             [[1, ""], [3, 4]],
         )
 
@@ -266,15 +394,13 @@ def test_rowcount(conn):
     cur.execute("select 1 from generate_series(1, 42)")
     assert cur.rowcount == 42
 
+    cur.execute("show timezone")
+    assert cur.rowcount == 1
+
     cur.execute("create table test_rowcount_notuples (id int primary key)")
     assert cur.rowcount == -1
 
-    cur.execute(
-        "insert into test_rowcount_notuples select generate_series(1, 42)"
-    )
-    assert cur.rowcount == 42
-
-    cur.close()
+    cur.execute("insert into test_rowcount_notuples select generate_series(1, 42)")
     assert cur.rowcount == 42
 
 
@@ -291,8 +417,9 @@ def test_rownumber(conn):
     assert cur.rownumber == 2
     cur.fetchmany(10)
     assert cur.rownumber == 12
-    rns = []
+    rns: List[int] = []
     for i in cur:
+        assert cur.rownumber
         rns.append(cur.rownumber)
         if len(rns) >= 3:
             break
@@ -418,6 +545,7 @@ def test_query_params_execute(conn):
     assert cur._query is None
 
     cur.execute("select %t, %s::text", [1, None])
+    assert cur._query is not None
     assert cur._query.query == b"select $1, $2::text"
     assert cur._query.params == [b"1", None]
 
@@ -473,10 +601,17 @@ def test_stream_row_factory(conn):
     assert next(it).a == 2
 
 
+def test_stream_no_row(conn):
+    cur = conn.cursor()
+    recs = list(cur.stream("select generate_series(2,1) as a"))
+    assert recs == []
+
+
+@pytest.mark.crdb_skip("no col query")
 def test_stream_no_col(conn):
     cur = conn.cursor()
-    it = iter(cur.stream("select"))
-    assert list(it) == [()]
+    recs = list(cur.stream("select"))
+    assert recs == [()]
 
 
 @pytest.mark.parametrize(
@@ -494,10 +629,61 @@ def test_stream_badquery(conn, query):
             pass
 
 
+def test_stream_error_tx(conn):
+    cur = conn.cursor()
+    with pytest.raises(psycopg.ProgrammingError):
+        for rec in cur.stream("wat"):
+            pass
+    assert conn.info.transaction_status == conn.TransactionStatus.INERROR
+
+
+def test_stream_error_notx(conn):
+    conn.autocommit = True
+    cur = conn.cursor()
+    with pytest.raises(psycopg.ProgrammingError):
+        for rec in cur.stream("wat"):
+            pass
+    assert conn.info.transaction_status == conn.TransactionStatus.IDLE
+
+
+def test_stream_error_python_to_consume(conn):
+    cur = conn.cursor()
+    with pytest.raises(ZeroDivisionError):
+        with closing(cur.stream("select generate_series(1, 10000)")) as gen:
+            for rec in gen:
+                1 / 0
+    assert conn.info.transaction_status in (
+        conn.TransactionStatus.INTRANS,
+        conn.TransactionStatus.INERROR,
+    )
+
+
+def test_stream_error_python_consumed(conn):
+    cur = conn.cursor()
+    with pytest.raises(ZeroDivisionError):
+        gen = cur.stream("select 1")
+        for rec in gen:
+            1 / 0
+    gen.close()
+    assert conn.info.transaction_status == conn.TransactionStatus.INTRANS
+
+
+def test_stream_close(conn):
+    cur = conn.cursor()
+    with pytest.raises(psycopg.OperationalError):
+        for rec in cur.stream("select generate_series(1, 3)"):
+            if rec[0] == 1:
+                conn.close()
+            else:
+                assert False
+
+    assert conn.closed
+
+
 def test_stream_binary_cursor(conn):
     cur = conn.cursor(binary=True)
     recs = []
-    for rec in cur.stream("select generate_series(1, 2)"):
+    for rec in cur.stream("select x::int4 from generate_series(1, 2) x"):
         recs.append(rec)
         assert cur.pgresult.fformat(0) == 1
         assert cur.pgresult.get_value(0, 0) == bytes([0, 0, 0, rec[0]])
@@ -508,7 +694,7 @@ def test_stream_binary_cursor(conn):
 def test_stream_execute_binary(conn):
     cur = conn.cursor()
     recs = []
-    for rec in cur.stream("select generate_series(1, 2)", binary=True):
+    for rec in cur.stream("select x::int4 from generate_series(1, 2) x", binary=True):
         recs.append(rec)
         assert cur.pgresult.fformat(0) == 1
         assert cur.pgresult.get_value(0, 0) == bytes([0, 0, 0, rec[0]])
@@ -570,7 +756,10 @@ class TestColumn:
         assert c.name == "now"
         assert c.type_code == builtins["date"].oid
         assert c.display_size is None
-        assert c.internal_size == 4
+        if is_crdb(conn):
+            assert c.internal_size == 16
+        else:
+            assert c.internal_size == 4
         assert c.precision is None
         assert c.scale is None
 
@@ -590,8 +779,8 @@ class TestColumn:
             ("numeric(10)", 10, 0, None, None),
             ("numeric(10, 3)", 10, 3, None, None),
             ("time", None, None, None, 8),
-            ("time(4)", 4, None, None, 8),
-            ("time(10)", 6, None, None, 8),
+            crdb_time_precision("time(4)", 4, None, None, 8),
+            crdb_time_precision("time(10)", 6, None, None, 8),
         ],
     )
     def test_details(self, conn, type, precision, scale, dsize, isize):
@@ -618,14 +807,40 @@ class TestColumn:
         unpickled = pickle.loads(pickled)
         assert [tuple(d) for d in description] == [tuple(d) for d in unpickled]
 
+    @pytest.mark.crdb_skip("no col query")
     def test_no_col_query(self, conn):
         cur = conn.execute("select")
         assert cur.description == []
         assert cur.fetchall() == [()]
 
+    def test_description_closed_connection(self, conn):
+        # If we have reasons to break this test we will (e.g. we really need
+        # the connection). In #172 it fails just by accident.
+        cur = conn.execute("select 1::int4 as foo")
+        conn.close()
+        assert len(cur.description) == 1
+        col = cur.description[0]
+        assert col.name == "foo"
+        assert col.type_code == 23
+
+    def test_name_not_a_name(self, conn):
+        cur = conn.cursor()
+        (res,) = cur.execute("""select 'x' as "foo-bar" """).fetchone()
+        assert res == "x"
+        assert cur.description[0].name == "foo-bar"
+
+    @pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
+    def test_name_encode(self, conn, encoding):
+        conn.execute(f"set client_encoding to {encoding}")
+        cur = conn.cursor()
+        (res,) = cur.execute("""select 'x' as "\u20ac" """).fetchone()
+        assert res == "x"
+        assert cur.description[0].name == "\u20ac"
+
 
 def test_str(conn):
     cur = conn.cursor()
+    assert "psycopg.Cursor" in str(cur)
     assert "[IDLE]" in str(cur)
     assert "[closed]" not in str(cur)
     assert "[no result]" in str(cur)
@@ -643,17 +858,15 @@ def test_str(conn):
 @pytest.mark.parametrize("fmt", PyFormat)
 @pytest.mark.parametrize("fmt_out", pq.Format)
 @pytest.mark.parametrize("fetch", ["one", "many", "all", "iter"])
-@pytest.mark.parametrize(
-    "row_factory", ["tuple_row", "dict_row", "namedtuple_row"]
-)
-def test_leak(dsn, faker, fmt, fmt_out, fetch, row_factory, retries):
+@pytest.mark.parametrize("row_factory", ["tuple_row", "dict_row", "namedtuple_row"])
+def test_leak(conn_cls, dsn, faker, fmt, fmt_out, fetch, row_factory):
     faker.format = fmt
     faker.choose_schema(ncols=5)
     faker.make_records(10)
     row_factory = getattr(rows, row_factory)
 
     def work():
-        with psycopg.connect(dsn) as conn:
+        with conn_cls.connect(dsn) as conn, conn.transaction(force_rollback=True):
             with conn.cursor(binary=fmt_out, row_factory=row_factory) as cur:
                 cur.execute(faker.drop_stmt)
                 cur.execute(faker.create_stmt)
@@ -663,12 +876,12 @@ def test_leak(dsn, faker, fmt, fmt_out, fetch, row_factory, retries):
                 cur.execute(faker.select_stmt)
 
                 if fetch == "one":
-                    while 1:
+                    while True:
                         tmp = cur.fetchone()
                         if tmp is None:
                             break
                 elif fetch == "many":
-                    while 1:
+                    while True:
                         tmp = cur.fetchmany(3)
                         if not tmp:
                             break
@@ -678,28 +891,23 @@ def test_leak(dsn, faker, fmt, fmt_out, fetch, row_factory, retries):
                     for rec in cur:
                         pass
 
-    for retry in retries:
-        with retry:
-            n = []
-            gc_collect()
-            for i in range(3):
-                work()
-                gc_collect()
-                n.append(len(gc.get_objects()))
-            assert (
-                n[0] == n[1] == n[2]
-            ), f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"
+    n = []
+    gc_collect()
+    for i in range(3):
+        work()
+        gc_collect()
+        n.append(len(gc.get_objects()))
+    assert n[0] == n[1] == n[2], f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"
 
 
-def my_row_factory(cursor):
+def my_row_factory(
+    cursor: Union[psycopg.Cursor[List[str]], psycopg.AsyncCursor[List[str]]]
+) -> RowMaker[List[str]]:
     if cursor.description is not None:
         titles = [c.name for c in cursor.description]
 
         def mkrow(values):
-            return [
-                f"{value.upper()}{title}"
-                for title, value in zip(titles, values)
-            ]
+            return [f"{value.upper()}{title}" for title, value in zip(titles, values)]
 
         return mkrow
     else:
