@@ -1,3 +1,4 @@
+import enum
 from decimal import Decimal
 from math import isnan, isinf, exp
 
@@ -9,6 +10,7 @@ from psycopg import sql
 from psycopg.adapt import Transformer, PyFormat
 from psycopg.types.numeric import FloatLoader
 
+from ..fix_crdb import is_crdb
 
 #
 # Tests with int
@@ -23,15 +25,15 @@ from psycopg.types.numeric import FloatLoader
         (-1, "'-1'::int"),
         (42, "'42'::smallint"),
         (-42, "'-42'::smallint"),
-        (int(2 ** 63 - 1), "'9223372036854775807'::bigint"),
-        (int(-(2 ** 63)), "'-9223372036854775808'::bigint"),
+        (int(2**63 - 1), "'9223372036854775807'::bigint"),
+        (int(-(2**63)), "'-9223372036854775808'::bigint"),
     ],
 )
 @pytest.mark.parametrize("fmt_in", PyFormat)
 def test_dump_int(conn, val, expr, fmt_in):
     assert isinstance(val, int)
     cur = conn.cursor()
-    cur.execute(f"select {expr} = %{fmt_in}", (val,))
+    cur.execute(f"select {expr} = %{fmt_in.value}", (val,))
     assert cur.fetchone()[0] is True
 
 
@@ -43,27 +45,27 @@ def test_dump_int(conn, val, expr, fmt_in):
         (-1, "'-1'::smallint"),
         (42, "'42'::smallint"),
         (-42, "'-42'::smallint"),
-        (int(2 ** 15 - 1), f"'{2 ** 15 - 1}'::smallint"),
-        (int(-(2 ** 15)), f"'{-2 ** 15}'::smallint"),
-        (int(2 ** 15), f"'{2 ** 15}'::integer"),
-        (int(-(2 ** 15) - 1), f"'{-2 ** 15 - 1}'::integer"),
-        (int(2 ** 31 - 1), f"'{2 ** 31 - 1}'::integer"),
-        (int(-(2 ** 31)), f"'{-2 ** 31}'::integer"),
-        (int(2 ** 31), f"'{2 ** 31}'::bigint"),
-        (int(-(2 ** 31) - 1), f"'{-2 ** 31 - 1}'::bigint"),
-        (int(2 ** 63 - 1), f"'{2 ** 63 - 1}'::bigint"),
-        (int(-(2 ** 63)), f"'{-2 ** 63}'::bigint"),
-        (int(2 ** 63), f"'{2 ** 63}'::numeric"),
-        (int(-(2 ** 63) - 1), f"'{-2 ** 63 - 1}'::numeric"),
+        (int(2**15 - 1), f"'{2 ** 15 - 1}'::smallint"),
+        (int(-(2**15)), f"'{-2 ** 15}'::smallint"),
+        (int(2**15), f"'{2 ** 15}'::integer"),
+        (int(-(2**15) - 1), f"'{-2 ** 15 - 1}'::integer"),
+        (int(2**31 - 1), f"'{2 ** 31 - 1}'::integer"),
+        (int(-(2**31)), f"'{-2 ** 31}'::integer"),
+        (int(2**31), f"'{2 ** 31}'::bigint"),
+        (int(-(2**31) - 1), f"'{-2 ** 31 - 1}'::bigint"),
+        (int(2**63 - 1), f"'{2 ** 63 - 1}'::bigint"),
+        (int(-(2**63)), f"'{-2 ** 63}'::bigint"),
+        (int(2**63), f"'{2 ** 63}'::numeric"),
+        (int(-(2**63) - 1), f"'{-2 ** 63 - 1}'::numeric"),
     ],
 )
 @pytest.mark.parametrize("fmt_in", PyFormat)
 def test_dump_int_subtypes(conn, val, expr, fmt_in):
     cur = conn.cursor()
-    cur.execute(f"select pg_typeof({expr}) = pg_typeof(%{fmt_in})", (val,))
+    cur.execute(f"select pg_typeof({expr}) = pg_typeof(%{fmt_in.value})", (val,))
     assert cur.fetchone()[0] is True
     cur.execute(
-        f"select {expr} = %(v){fmt_in}, {expr}::text, %(v){fmt_in}::text",
+        f"select {expr} = %(v){fmt_in.value}, {expr}::text, %(v){fmt_in.value}::text",
         {"v": val},
     )
     ok, want, got = cur.fetchone()
@@ -71,16 +73,21 @@ def test_dump_int_subtypes(conn, val, expr, fmt_in):
     assert ok
 
 
+class MyEnum(enum.IntEnum):
+    foo = 42
+
+
+class MyMixinEnum(enum.IntEnum):
+    foo = 42000000
+
+
 @pytest.mark.parametrize("fmt_in", PyFormat)
-def test_dump_enum(conn, fmt_in):
-    import enum
-
-    class MyEnum(enum.IntEnum):
-        foo = 42
-
+@pytest.mark.parametrize("enum", [MyEnum, MyMixinEnum])
+def test_dump_enum(conn, fmt_in, enum):
     cur = conn.cursor()
-    (res,) = cur.execute("select %s", (MyEnum.foo,)).fetchone()
-    assert res == 42
+    cur.execute(f"select %{fmt_in.value}", (enum.foo,))
+    (res,) = cur.fetchone()
+    assert res == enum.foo.value
 
 
 @pytest.mark.parametrize(
@@ -91,12 +98,12 @@ def test_dump_enum(conn, fmt_in):
         (-1, b" -1"),
         (42, b"42"),
         (-42, b" -42"),
-        (int(2 ** 63 - 1), b"9223372036854775807"),
-        (int(-(2 ** 63)), b" -9223372036854775808"),
-        (int(2 ** 63), b"9223372036854775808"),
-        (int(-(2 ** 63 + 1)), b" -9223372036854775809"),
-        (int(2 ** 100), b"1267650600228229401496703205376"),
-        (int(-(2 ** 100)), b" -1267650600228229401496703205376"),
+        (int(2**63 - 1), b"9223372036854775807"),
+        (int(-(2**63)), b" -9223372036854775808"),
+        (int(2**63), b"9223372036854775808"),
+        (int(-(2**63 + 1)), b" -9223372036854775809"),
+        (int(2**100), b"1267650600228229401496703205376"),
+        (int(-(2**100)), b" -1267650600228229401496703205376"),
     ],
 )
 def test_quote_int(conn, val, expr):
@@ -131,6 +138,8 @@ def test_quote_int(conn, val, expr):
 )
 @pytest.mark.parametrize("fmt_out", pq.Format)
 def test_load_int(conn, val, pgtype, want, fmt_out):
+    if pgtype == "integer" and is_crdb(conn):
+        pgtype = "int4"  # "integer" is "int8" on crdb
     cur = conn.cursor(binary=fmt_out)
     cur.execute(f"select %s::{pgtype}", (val,))
     assert cur.pgresult.fformat(0) == fmt_out
@@ -168,7 +177,7 @@ def test_load_int(conn, val, pgtype, want, fmt_out):
 def test_dump_float(conn, val, expr, fmt_in):
     assert isinstance(val, float)
     cur = conn.cursor()
-    cur.execute(f"select %{fmt_in} = {expr}::float8", (val,))
+    cur.execute(f"select %{fmt_in.value} = {expr}::float8", (val,))
     assert cur.fetchone()[0] is True
 
 
@@ -205,8 +214,8 @@ def test_quote_float(conn, val, expr):
 @pytest.mark.parametrize(
     "val, expr",
     [
-        (exp(1), "exp(1)"),
-        (-exp(1), "-exp(1)"),
+        (exp(1), "exp(1.0)"),
+        (-exp(1), "-exp(1.0)"),
         (1e30, "'1e30'"),
         (1e-30, "1e-30"),
         (-1e30, "'-1e30'"),
@@ -216,14 +225,10 @@ def test_quote_float(conn, val, expr):
 def test_dump_float_approx(conn, val, expr):
     assert isinstance(val, float)
     cur = conn.cursor()
-    cur.execute(
-        f"select abs(({expr}::float8 - %s) / {expr}::float8) <= 1e-15", (val,)
-    )
+    cur.execute(f"select abs(({expr}::float8 - %s) / {expr}::float8) <= 1e-15", (val,))
     assert cur.fetchone()[0] is True
 
-    cur.execute(
-        f"select abs(({expr}::float4 - %s) / {expr}::float4) <= 1e-6", (val,)
-    )
+    cur.execute(f"select abs(({expr}::float4 - %s) / {expr}::float4) <= 1e-6", (val,))
     assert cur.fetchone()[0] is True
 
 
@@ -277,10 +282,10 @@ def test_load_float(conn, val, pgtype, want, fmt_out):
 @pytest.mark.parametrize(
     "expr, pgtype, want",
     [
-        ("exp(1)", "float4", 2.71828),
-        ("-exp(1)", "float4", -2.71828),
-        ("exp(1)", "float8", 2.71828182845905),
-        ("-exp(1)", "float8", -2.71828182845905),
+        ("exp(1.0)", "float4", 2.71828),
+        ("-exp(1.0)", "float4", -2.71828),
+        ("exp(1.0)", "float8", 2.71828182845905),
+        ("-exp(1.0)", "float8", -2.71828182845905),
         ("1.42e10", "float4", 1.42e10),
         ("-1.42e10", "float4", -1.42e10),
         ("1.42e40", "float8", 1.42e40),
@@ -296,6 +301,7 @@ def test_load_float_approx(conn, expr, pgtype, want, fmt_out):
     assert result == pytest.approx(want)
 
 
+@pytest.mark.crdb_skip("copy")
 def test_load_float_copy(conn):
     cur = conn.cursor(binary=False)
     with cur.copy("copy (select 3.14::float8, 'hi'::text) to stdout;") as copy:
@@ -328,7 +334,7 @@ def test_load_float_copy(conn):
 def test_roundtrip_numeric(conn, val, fmt_in, fmt_out):
     cur = conn.cursor(binary=fmt_out)
     val = Decimal(val)
-    cur.execute(f"select %{fmt_in}", (val,))
+    cur.execute(f"select %{fmt_in.value}", (val,))
     result = cur.fetchone()[0]
     assert isinstance(result, Decimal)
     if val.is_nan():
@@ -363,6 +369,7 @@ def test_quote_numeric(conn, val, expr):
         assert r == (val, -val)
 
 
+@pytest.mark.crdb_skip("binary decimal")
 @pytest.mark.parametrize(
     "expr",
     ["NaN", "1", "1.0", "-1", "0.0", "0.01", "11", "1.1", "1.01", "0", "0.00"]
@@ -392,7 +399,15 @@ def test_dump_numeric_binary(conn, expr):
 
 
 @pytest.mark.slow
-@pytest.mark.parametrize("fmt_in", PyFormat)
+@pytest.mark.parametrize(
+    "fmt_in",
+    [
+        f
+        if f != PyFormat.BINARY
+        else pytest.param(f, marks=pytest.mark.crdb_skip("binary decimal"))
+        for f in PyFormat
+    ],
+)
 def test_dump_numeric_exhaustive(conn, fmt_in):
     cur = conn.cursor()
 
@@ -417,10 +432,8 @@ def test_dump_numeric_exhaustive(conn, fmt_in):
         for f in funcs:
             expr = f(i)
             val = Decimal(expr)
-            cur.execute(
-                f"select %{fmt_in}::text, %s::decimal::text", [val, expr]
-            )
-            want, got = cur.fetchone()
+            cur.execute(f"select %{fmt_in.value}::text, %s::decimal::text", [val, expr])
+            got, want = cur.fetchone()
             assert got == want
 
 
@@ -569,6 +582,18 @@ def test_minus_minus_quote(conn, pgtype):
     assert result == 1
 
 
+@pytest.mark.parametrize("wrapper", "Int2 Int4 Int8 Oid Float4 Float8".split())
+@pytest.mark.parametrize("fmt_in", PyFormat)
+def test_dump_wrapper(conn, wrapper, fmt_in):
+    wrapper = getattr(psycopg.types.numeric, wrapper)
+    obj = wrapper(1)
+    cur = conn.execute(
+        f"select %(obj){fmt_in.value} = 1, %(obj){fmt_in.value}", {"obj": obj}
+    )
+    rec = cur.fetchone()
+    assert rec[0], rec[1]
+
+
 @pytest.mark.parametrize("wrapper", "Int2 Int4 Int8 Oid Float4 Float8".split())
 def test_dump_wrapper_oid(wrapper):
     wrapper = getattr(psycopg.types.numeric, wrapper)
@@ -579,10 +604,11 @@ def test_dump_wrapper_oid(wrapper):
     assert repr(wrapper(n)) == f"{wrapper.__name__}({n})"
 
 
+@pytest.mark.crdb("skip", reason="all types returned as bigint? TODOCRDB")
 @pytest.mark.parametrize("wrapper", "Int2 Int4 Int8 Oid Float4 Float8".split())
 @pytest.mark.parametrize("fmt_in", PyFormat)
 def test_repr_wrapper(conn, wrapper, fmt_in):
     wrapper = getattr(psycopg.types.numeric, wrapper)
-    cur = conn.execute(f"select pg_typeof(%{fmt_in})::oid", [wrapper(0)])
+    cur = conn.execute(f"select pg_typeof(%{fmt_in.value})::oid", [wrapper(0)])
     oid = cur.fetchone()[0]
     assert oid == psycopg.postgres.types[wrapper.__name__.lower()].oid
